;
; KaKeeware cmdTotal is yet another upgrade to IETOTALX proggie.
; http://www.kakeeware.com/i_ietotalx.php
;
; This is a generic approach to make it possible to use any
; Total Commander WCX packer plugins from a command line.
;
; -- How to use? --
; From a command line, run:
;    cmdTotal [wcxPluginName] [option] [fileName] [targetDirectory]
;     where:
;       [wcxPluginName] - WCX plugin filename (path not needed if in the same directory)
;       [option] - can be:
;         l - to list the files
;         t - to test the files
;         x - to extract the files
;       [fileName] - file you want to process
;       [targetDirectory] - where to extract the files (default: _[fileName])
;
; Examples:
;     cmdTotal.exe InstExpl.wcx l WinPcap_3_1.exe             - to list the files inside the archive
;     cmdTotal.exe InstExpl.wcx x WinPcap_3_1.exe             - to extract the files inside the archive to _WinPcap_3_1.exe directory
;     cmdTotal.exe InstExpl.wcx x WinPcap_3_1.exe c:\unpacked - to extract the files inside the archive to c:\unpacked directory
;
; -- Technical note --
; There's a guard thread created at the beginning of the main routine that will kill the instance
; of cmdTotal, if the main thread haven't finished its work after 300 seconds.
;
; There's SEH handler added before calling the function that extracts the files to make sure
; we extract as many of them as possible.
;
; -- History --
;  2007-01-03 v. 1.02 - fixed handling command line arguments (just made buffers a bit longer)
;  2006-12-31 v. 1.01 - added support for SetProcessDataProc (some plugins expect callback to be set) - (thx to perchwasamara for hint on that)
;                     - fixed the way ProcessFile is called - some plugins I tested use the arguments provided, some don't
;                     - got amazed couple of times by misleading information provided in a WCX plug-in HELP file, but luckily had OllyDbg at hand :) 
;  2006-12-30 v. 1.00 - first version
;
; -- Credits --
; - perchwasamara for hint on SetProcessDataProc
; - Jared B. for hint on problem with command line arguments
; - Hutch for his pieces of code I used in a proggie
; ;
; Feel free to play with the code to modify it up to your needs.
; If you happen to use my code, don't forget about creditz :-)
;
; Author: Adam Blaszczyk (c) 2006/2007
; WWW:    http://www.kakeeware.com
; e-mail: adam[]kakeeware[]com
;
; ============================================================


.586
.MODEL FLAT,STDCALL
 assume fs:nothing

 include windows.inc

 NULL  = 0
 CR    = 0Dh
 LF    = 0Ah

 INV  equ INVOKE
 OFS  equ OFFSET
 BPTR equ BYTE PTR
 WPTR equ WORD PTR
 DPTR equ DWORD PTR

 MOM  MACRO t:REQ, s:REQ
    push   s
    pop    t
 ENDM

 INCLUDEX  MACRO plik:REQ
     include    plik.inc
     includelib plik.lib
 ENDM

 INCX MACRO mods:VARARG
   FOR c,<mods>
    INCLUDEX c
   ENDM
 ENDM

 INCX    kernel32,user32,masm32

  PK_OM_LIST	   = 0
  PK_OM_EXTRACT = 1

  PK_SKIP	      = 0
  PK_TEST	      = 1
  PK_EXTRACT	   = 2

  OPENARCHIVEDATA struct
    oadArcName    dd ?
    oadOpenMode   dd ?
    oadOpenResult dd ?
    oadCmtBuf     dd ?
    oadCmtBufSize dd ?
    oadCmtSize    dd ?
    oadCmtState   dd ?
  OPENARCHIVEDATA ends

  HEADERDATA struct
    hdArcName     db 260 dup(?)
    hdFileName    db 260 dup(?)
    hdFlags       dd ?
    hdPackSize    dd ?
    hdUnpSize     dd ?
    hdHostOS      dd ?
    hdFileCRC     dd ?
    hdFileTime    dd ?
    hdUnpVer      dd ?
    hdMethod      dd ?
    hdFileAttr    dd ?
    hdCmtBuf      dd ?
    hdCmtBufSize  dd ?
    hdCmtSize     dd ?
    hdCmtState    dd ?
  HEADERDATA ends

.data?
 curDirName          db 1024 dup(?)
 sBuf                db 1024 dup(?)
 argPluginName       db 1024 dup(?)
 argOption           db 1024 dup(?)
 argFileName         db 1024 dup(?)
 argTargetDir        db 1024 dup(?)
 hStdOut             dd ?
 hFile               dd ?
 openMode            dd ?
 processMode         dd ?
 libHandle           dd ?
 OpenArchive         dd ?
 ReadHeader          dd ?
 ProcessFile         dd ?
 CloseArchive        dd ?
 SetProcessDataProc  dd ?
 archiveData         OPENARCHIVEDATA <>
 archiveHandle       dd ?
 headerData          HEADERDATA <>

 PreservedEip        dd ?
 PreservedEsp        dd ?
 PreservedEbp        dd ?
.data
  szInfo                db CR,LF,'cmdTotal 1.02 (c) 2006/2007 KaKeeware, http://www.kakeeware.com',CR,LF
                        db NULL
  szHelp                db CR,LF,'Usage: cmdTotal [wcxPluginName] [option] [fileName] [targetDirectory]',CR,LF
                        db '   where:',CR,LF
                        db '     [wcxPluginName] - WCX plugin filename (must be in the same directory)',CR,LF
                        db '     [option] - can be:',CR,LF
                        db '       l - to list the files',CR,LF
                        db '       t - to test the files',CR,LF
                        db '       x - to extract the files',CR,LF
                        db '     [fileName] - file you want to process',CR,LF
                        db '     [targetDirectory] - where to extract the files (default: _[fileName])',CR,LF
                        db 'Credits: perchwasamara, Jared B, Hutch',CR,LF
                        db CR,LF,NULL
  szProcessing          db CR,LF,'Using plugin:     %s',CR,LF
                        db CR,LF,'Using option:     %s',CR,LF
                        db CR,LF,'Processing file:  %s',CR,LF
                        db CR,LF,'Target directory: %s',CR,LF
                        db NULL
  oneFile               db ' --> %s',CR,LF,NULL
  insideFile            db '    -> %s',CR,LF,NULL
  backslash             db '\',NULL
  backslashUnderscore   db '\_',NULL

  szOpenArchive         db 'OpenArchive',NULL
  szReadHeader          db 'ReadHeader',NULL
  szProcessFile         db 'ProcessFile',NULL
  szCloseArchive        db 'CloseArchive',NULL
  szSetProcessDataProc  db 'SetProcessDataProc',NULL

  szGiveMeAFileName     db 'Give me some command line parameters!',CR,LF,NULL
  szCantLoadLib         db 'Can',39,'t load the plug-in!',CR,LF,NULL
  szCantGetFunctions    db 'Can',39,'t retrieve addresses to plug-in functions!',CR,LF,NULL
  szCrash               db 'Sorry, exception occured -> quitting!',CR,LF,NULL
  szGuard               db 'Sorry, program didn',39,'t stop after 300 seconds -> quitting!',CR,LF,NULL

.code
 ; -----------------------------------------------------
 ;  Copyright notice:
 ; 'GetCLEx' function is almost an exact copy of GetCL function from masm32 package
 ; the only difference is the size of the local buffers - the sizes have been expanded to 1024 bytes
 ; -----------------------------------------------------
 GetCLEx proc ArgNum:DWORD, ItemBuffer:DWORD
 
   ; -------------------------------------------------
   ; arguments returned in "ItemBuffer"
   ;
   ; arg 0 = program name
   ; arg 1 = 1st arg
   ; arg 2 = 2nd arg etc....
   ; -------------------------------------------------
   ; Return values in eax
   ;
   ; 1 = successful operation
   ; 2 = no argument exists at specified arg number
   ; 3 = non matching quotation marks
   ; 4 = empty quotation marks
   ; -------------------------------------------------
 
     LOCAL lpCmdLine      :DWORD
     LOCAL cmdBuffer[1024] :BYTE
     LOCAL tmpBuffer[1024] :BYTE
 
     push esi
     push edi
 
     invoke GetCommandLine
     mov lpCmdLine, eax        ; address command line
 
   ; -------------------------------------------------
   ; count quotation marks to see if pairs are matched
   ; -------------------------------------------------
     xor ecx, ecx            ; zero ecx & use as counter
     mov esi, lpCmdLine
     
     @@:
       lodsb
       cmp al, 0
       je @F
       cmp al, 34            ; [ " ] character
       jne @B
       inc ecx               ; increment counter
       jmp @B
     @@:
 
     push ecx                ; save count
 
     shr ecx, 1              ; integer divide ecx by 2
     shl ecx, 1              ; multiply ecx by 2 to get dividend
 
     pop eax                 ; put count in eax
     cmp eax, ecx            ; check if they are the same
     je @F
       pop edi
       pop esi
       mov eax, 3            ; return 3 in eax = non matching quotation marks
       ret
     @@:
 
   ; ------------------------
   ; replace tabs with spaces
   ; ------------------------
     mov esi, lpCmdLine
     lea edi, cmdBuffer
 
     @@:
       lodsb
       cmp al, 0
       je rtOut
       cmp al, 9     ; tab
       jne rtIn
       mov al, 32
     rtIn:
       stosb
       jmp @B
     rtOut:
       stosb         ; write last byte
 
   ; -----------------------------------------------------------
   ; substitute spaces in quoted text with replacement character
   ; -----------------------------------------------------------
     lea eax, cmdBuffer
     mov esi, eax
     mov edi, eax
 
     subSt:
       lodsb
       cmp al, 0
       jne @F
       jmp subOut
     @@:
       cmp al, 34
       jne subNxt
       stosb
       jmp subSl     ; goto subloop
     subNxt:
       stosb
       jmp subSt
 
     subSl:
       lodsb
       cmp al, 32    ; space
       jne @F
         mov al, 254 ; substitute character
       @@:
       cmp al, 34
       jne @F
         stosb
         jmp subSt
       @@:
       stosb
       jmp subSl
 
     subOut:
       stosb         ; write last byte
 
   ; ----------------------------------------------------
   ; the following code determines the correct arg number
   ; and writes the arg into the destination buffer
   ; ----------------------------------------------------
     lea eax, cmdBuffer
     mov esi, eax
     lea edi, tmpBuffer
 
     mov ecx, 0          ; use ecx as counter
 
   ; ---------------------------
   ; strip leading spaces if any
   ; ---------------------------
     @@:
       lodsb
       cmp al, 32
       je @B
 
     l2St:
       cmp ecx, ArgNum     ; the number of the required cmdline arg
       je clSubLp2
       lodsb
       cmp al, 0
       je cl2Out
       cmp al, 32
       jne cl2Ovr           ; if not space
 
     @@:
       lodsb
       cmp al, 32          ; catch consecutive spaces
       je @B
 
       inc ecx             ; increment arg count
       cmp al, 0
       je cl2Out
 
     cl2Ovr:
       jmp l2St
 
     clSubLp2:
       stosb
     @@:
       lodsb
       cmp al, 32
       je cl2Out
       cmp al, 0
       je cl2Out
       stosb
       jmp @B
 
     cl2Out:
       mov al, 0
       stosb
 
   ; ------------------------------
   ; exit if arg number not reached
   ; ------------------------------
     .if ecx < ArgNum
       mov edi, ItemBuffer
       mov al, 0
       stosb
       mov eax, 2  ; return value of 2 means arg did not exist
       pop edi
       pop esi
       ret
     .endif
 
   ; -------------------------------------------------------------
   ; remove quotation marks and replace the substitution character
   ; -------------------------------------------------------------
     lea eax, tmpBuffer
     mov esi, eax
     mov edi, ItemBuffer
 
     rqStart:
       lodsb
       cmp al, 0
       je rqOut
       cmp al, 34    ; dont write [ " ] mark
       je rqStart
       cmp al, 254
       jne @F
       mov al, 32    ; substitute space
     @@:
       stosb
       jmp rqStart
 
   rqOut:
       stosb         ; write zero terminator
 
   ; ------------------
   ; handle empty quote
   ; ------------------
     mov esi, ItemBuffer
     lodsb
     cmp al, 0
     jne @F
     pop edi
     pop esi
     mov eax, 4  ; return value for empty quote
     ret
   @@:
 
     mov eax, 1  ; return value success
 
     pop edi
     pop esi
 
     ret
 
 GetCLEx endp
 
 ; -----------------------------------------------------
 ;  Copyright notice:
 ; 'print' function is based on StdOut function from masm32 package
 ; I rearranged it and removed pieces unnecessary in this proggie
 ; -----------------------------------------------------
 ; 
 print proc txt:DWORD
    LOCAL tmp:DWORD
    INV lstrlenA,txt
    xchg eax,edx
    INV WriteFile,hStdOut,txt,edx,ADDR tmp,NULL
    ret
 print endp

 fatalError proc
   INV print,OFS szCrash
   xor eax, eax
   inc eax       ; eax=1
   ret
 fatalError endp

 guardThread proc lPa:DWORD
   INV Sleep,300000 ; guard thread will close the program after 300 seconds
   INV print,OFS szGuard
   INV ExitProcess,-1
   ret
 guardThread endp

 sehit proc C exc,fra,con,dis:DWORD
   	mov eax, con
    push PreservedEip
    pop  [eax+CONTEXT.regEip]

    push PreservedEsp
    pop  [eax+CONTEXT.regEsp]

    push PreservedEbp
    pop  [eax+CONTEXT.regEbp]

    xor eax, eax	; ExceptionContinueExecution
    ret
  sehit endp
  
  ProcessDataProc proc fil,siz:DWORD
    mov al,1 ; return NON-ZERO     
    ret
  ProcessDataProc endp

Start:
 INV GetStdHandle,STD_OUTPUT_HANDLE
 mov hStdOut, eax

 INV SetUnhandledExceptionFilter,OFS fatalError

 INV CreateThread,NULL,NULL,OFS guardThread,NULL,NULL,NULL

 INV print,OFS szInfo

 INV GetCLEx,1,OFS argPluginName
 cmp al,1
 jne errBadCmdLine

 INV GetCLEx,2,OFS argOption
 cmp al,1
 jne errBadCmdLine

 mov al,BPTR argOption
 or  al,20h
 .if al=='l'
    mov openMode,PK_OM_LIST
    mov processMode,PK_SKIP
 .elseif al=='t'
    mov openMode,PK_OM_LIST
    mov processMode,PK_TEST
 .elseif al=='x'
    mov openMode,PK_OM_EXTRACT
    mov processMode,PK_EXTRACT
 .else
    jmp errBadCmdLine
 .endif

 INV GetCLEx,3,OFS argFileName
 cmp al,1
 jne errBadCmdLine

 .if openMode==PK_OM_EXTRACT
     INV GetCLEx,4,OFS argTargetDir
     .if eax==1
        ; create dir as requested
        INV lstrcpyA,OFS curDirName,OFS argTargetDir
     .else
        ; create underscoredDirectory directory
        INV GetCurrentDirectoryA,SIZEOF curDirName,OFS curDirName
        INV lstrcatA,OFS curDirName,OFS backslashUnderscore
        INV lstrcatA,OFS curDirName,OFS argFileName
        INV lstrcatA,OFS curDirName,OFS backslash
     .endif
     INV CreateDirectory,OFS curDirName,0
 .endif

 INV wsprintf,OFS sBuf,OFS szProcessing,OFS argPluginName,OFS argOption,OFS argFileName,OFS argTargetDir
 INV print,OFS sBuf

 ; load Plug-in library
 INV LoadLibrary,OFS argPluginName
 or  eax,eax
 je  errCantLoadLib

 mov libHandle,eax

 ; get offsets to exported functions
 INV GetProcAddress,libHandle,OFS szOpenArchive
 or  eax,eax
 je  errCantGetFunctions
 mov OpenArchive,eax

 INV GetProcAddress,libHandle,OFS szReadHeader
 or  eax,eax
 je  errCantGetFunctions
 mov ReadHeader,eax

 INV GetProcAddress,libHandle,OFS szProcessFile
 or  eax,eax
 je  errCantGetFunctions
 mov ProcessFile,eax

 INV GetProcAddress,libHandle,OFS szCloseArchive
 or  eax,eax
 je  errCantGetFunctions
 mov CloseArchive,eax

 ; this function is not mandatory, so we don't care if it is exported
 ; although we will try to call it to set our callback
 ; - just in case the plug-in implements it and it assumes 
 ; it will be set to something <> NULL
 ; (thx to perchwasamara for hint on that)
 INV GetProcAddress,libHandle,OFS szSetProcessDataProc
 mov SetProcessDataProc,eax

 ; we are going to open the archive for extracting
  mov archiveData.oadArcName,OFS argFileName
  MOM archiveData.oadOpenMode,openMode
  xor eax,eax
  mov archiveData.oadCmtBuf,eax
  mov archiveData.oadCmtBufSize,eax
  mov archiveData.oadCmtSize,eax
  mov archiveData.oadCmtState,eax
  push OFS archiveData
  call DPTR [OpenArchive]
  .if eax!=0
       mov  archiveHandle,eax
       
       .if SetProcessDataProc!=0          
          push ProcessDataProc
          push archiveHandle
          call DPTR [SetProcessDataProc]
       .endif
     nextHeader:
        push OFS headerData
        push archiveHandle
        call DPTR [ReadHeader]
        .if eax==0
            INV wsprintfA,OFS sBuf,OFS insideFile,OFS headerData.hdFileName
            INV print,OFS sBuf

            push OFS sehit ; setup SEH frame to handle potential exceptions
            push fs:[0]
            mov  fs:[0], esp
            mov  PreservedEbp,ebp
            mov  PreservedEsp,esp
            mov  PreservedEip,OFS welcomeBack

            push OFS headerData.hdFileName
            push OFS curDirName
            push processMode
            push archiveHandle
            call DPTR [ProcessFile]

     welcomeBack:
            pop fs:[0] ; - restore SEH
            add esp,4  ; /

            jmp nextHeader
        .endif

        ; close the archive
        push archiveHandle
        call DPTR [CloseArchive]
  .endif

  INV FreeLibrary,libHandle

  jmp done

 errCantGetFunctions:
  INV  print,OFS szCantGetFunctions
  jmp  done

 errCantLoadLib:
  INV  print,OFS szCantLoadLib
  jmp  done

 errBadCmdLine:
  INV print,OFS szHelp
  INV print,OFS szGiveMeAFileName

done:
  INV ExitProcess,0

END Start